package dk.apaq.framework.repository.vaadin;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import com.vaadin.data.Buffered;
import com.vaadin.data.Container;
import com.vaadin.data.Item;
import com.vaadin.data.Property;
import com.vaadin.data.Validator.InvalidValueException;
import dk.apaq.framework.criteria.Criteria;
import dk.apaq.framework.criteria.Sorter;
import dk.apaq.framework.repository.BaseRepositoryListener;
import dk.apaq.framework.repository.Repository;
import dk.apaq.framework.repository.RepositoryEvent.WithId;
import dk.apaq.framework.repository.RepositoryEvent.WithIdAndEntity;
import dk.apaq.framework.repository.RepositoryNotifier;

/**
 * A Vaadin Container that uses a <code>Repository</code> as a backend. The container supports the operations
 * supported by its crud while caching the object it reads in a WeakHashMap.
 *
 * - Crud.Editable -
 * If the Crud is a <code>Crud.Editable</code> instance, then it will be possible to use the addItem() and
 * removeItem() methods of the container.
 *
 * - Crud.Filterable -
 * If the Crud is an instance of <code>Crud.Filterable</code> then it will be possible to specifiy <code>Filter</code>
 * and <code>Sorter</code> on the container to have its data filtered directly in the Crud backend.
 *
 * - CrudNotifier -
 * If the Crud is an instance of <code>CrudNotifer</code> then the container will make sure to invalidate
 * changed items in the cache and reload idlist if needed.
 *
 */
public class RepositoryContainer<BT, IDT> implements Container, Container.Indexed, Container.Ordered, Container.ItemSetChangeNotifier, Container.Sortable, Buffered {

    protected Criteria criteria;
    protected final Repository<BT, IDT> repository;
    protected List<IDT> idList = null;
    protected transient LinkedHashMap<String, Class> model;
    private final Map<IDT, RepositoryItem> itemIdToItem = new WeakHashMap<IDT, RepositoryItem>();
    private List<ItemSetChangeListener> itemSetChangeListeners = new ArrayList<ItemSetChangeListener>();
    private CrudChangeHandler repositoryChangeHandler = null;
    private boolean readThrough = true;
    private boolean writeThrough = false;
    private final Class beanClass;
    private Object lastItemIdPropertyRetrieved = null;
    private Item lastItemPropertyRetrieved = null;

    public void sort(Object[] propertyId, boolean[] ascending) {
        List<Sorter.SorterEntry> sorterEntries = new ArrayList<Sorter.SorterEntry>();

        for(int i=0; i<propertyId.length; i++) {
            String id = propertyId[i].toString();
            Sorter.Direction direction = ascending[i] ? Sorter.Direction.Ascending : Sorter.Direction.Descending;
            sorterEntries.add(new Sorter.SorterEntry(id, direction));
        }
        
        Sorter sorter = new Sorter(sorterEntries);
        
        Criteria c = criteria == null ? new Criteria(sorter) : new Criteria(criteria.getRule(), sorter, criteria.getLimit());
        setCriteria(c);
    }

    @Override
    public Collection<?> getSortableContainerPropertyIds() {
        return getContainerPropertyIds();
    }
    
    private class CrudChangeHandler extends BaseRepositoryListener<BT, IDT> {

        @Override
        public void onEntityCreate(WithIdAndEntity<BT, IDT> event) {
            updateIds();
        }

        @Override
        public void onEntityUpdate(WithIdAndEntity<BT, IDT> event) {
            invalidate(event.getEntityId());
            updateIds();
        }

        @Override
        public void onEntityDelete(WithId<BT, IDT> event) {
            if (idList != null && idList.contains(event.getEntityId())) {
                updateIds();
            }
        }
    }

    private class ChangeEvent implements ItemSetChangeEvent {

        private final Container parent;

        public ChangeEvent(Container parent) {
            this.parent = parent;
        }

        @Override
        public Container getContainer() {
            return parent;
        }
    }

    
    public RepositoryContainer(Repository<BT, IDT> repository, Class beanClass) {
        this(repository, beanClass, getPropertyDescriptors(beanClass));
    }

    public RepositoryContainer(Repository<BT, IDT> repository, Class beanClass, boolean excludeCollectionProperties) {
        this(repository, beanClass, getPropertyDescriptors(beanClass, excludeCollectionProperties));
    }

    protected RepositoryContainer(Repository<BT, IDT> repository, Class beanClass, LinkedHashMap<String, Class> propertyMap) {
        this.repository = repository;
        this.model = propertyMap;
        this.beanClass = beanClass;
        
        if (this.repository instanceof RepositoryNotifier) {
            repositoryChangeHandler = new CrudChangeHandler();
            ((RepositoryNotifier) this.repository).addListener(repositoryChangeHandler);
        }
    }

    protected RepositoryItem buildItem(RepositoryContainer<BT, IDT> container, Repository<BT, IDT> crud, IDT id, BT bean) {
        RepositoryItem ri = new RepositoryItem(container, crud, id, bean);
        ri.setWriteThrough(writeThrough);
        return ri;
    }

    @Override
    public Item getItem(Object itemId) {
        if (itemId == null) {
            return null;
        }

        RepositoryItem item = null;
        
        if(!readThrough) {
            item = itemIdToItem.get((IDT) itemId);
        }
        
        if (item == null) {
            BT bean = repository.findOne((IDT) itemId);
            if (bean == null) {
                return null;
            }
            item = buildItem(this, repository, (IDT) itemId, bean);

            if(!readThrough) {
                itemIdToItem.put((IDT) itemId, item);
            }
        }
        return item;
    }

    public Criteria getCriteria() {
        return criteria;
    }
    
    public void setCriteria(Criteria criteria) {
        this.criteria = criteria;
        updateIds();
    }

    /**
     * Reloads the list of ids matching the specified filter and sorter.
     */
    protected void updateIds() {
        boolean firstUpdate = idList == null;
        idList = getItemIdList(criteria);
        
        if(!firstUpdate) {
            fireItemSetChange();
        }
    }

    /**
     * Loads list of ids from the crud.
     */
    protected List<IDT> getItemIdList(Criteria criteria) {
        if (criteria != null) {
            return repository.findAllIds(criteria);
        } else {
            return repository.findAllIds();
        }
    }

    /**
     * Invalidates a specific item.
     */
    protected void invalidate(IDT id) {
        itemIdToItem.remove(id);
        /*CrudItem<IDT, BT> item = itemIdToItem.get(id);
        if (item != null && !item.isModified()) {
            item.setInnerBean(crud.read(id));
        }*/
    }

    @Override
    public void addListener(ItemSetChangeListener listener) {
        itemSetChangeListeners.add(listener);
    }

    @Override
    public void removeListener(ItemSetChangeListener listener) {
        itemSetChangeListeners.remove(listener);
    }

    /**
     * Tells listeners that set of items in the container has changed.
     */
    private void fireItemSetChange() {
        ChangeEvent event = new ChangeEvent(this);
        for (ItemSetChangeListener listener : itemSetChangeListeners) {
            listener.containerItemSetChange(event);
        }
    }

    @Override
    public boolean addContainerProperty(Object propertyId, Class<?> type, Object defaultValue) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported by CrudContainers");
    }

    @Override
    public Object addItem() throws UnsupportedOperationException {
        try {
            BT entity = repository.save((BT)beanClass.newInstance());
            updateIds();
            return repository.getIdFromEntity(entity);
        } catch (InstantiationException ex) {
            throw new UnsupportedOperationException("Unable to create instance.", ex);
        } catch (IllegalAccessException ex) {
            throw new UnsupportedOperationException("Unable to create instance.", ex);
        }
        
    }

    @Override
    public Item addItem(Object itemId) throws UnsupportedOperationException {
        return null;
    }

    @Override
    public Object addItemAfter(Object previousItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Item addItemAfter(Object previousItemId, Object newItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object addItemAt(int index) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Item addItemAt(int index, Object newItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean containsId(Object itemId) {
        return getItemIdsAsList().contains(itemId);
    }

    @Override
    public Object firstItemId() {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return null;
        } else {
            return idlist.get(0);
        }
    }

    @Override
    public Object getIdByIndex(int index) {
        return getItemIdsAsList().get(index);
    }

    @Override
    public Collection<?> getItemIds() {
        return Collections.unmodifiableCollection(getItemIdsAsList());
    }

    public List<?> getItemIdsAsList() {
        if (idList == null) {
            updateIds();
        }
        return idList;
    }

    @Override
    public Collection<?> getContainerPropertyIds() {
        return model.keySet();
    }

    @Override
    public Property getContainerProperty(Object itemId, Object propertyId) {
        
        Item item;
        
        //Check if the itemid is the same as the one we have retrieved last
        //if it is then resuse the last item we got.
        if(itemId.equals(lastItemIdPropertyRetrieved)) {
            item = lastItemPropertyRetrieved;
        } else {
            item = getItem(itemId);
            lastItemIdPropertyRetrieved = itemId;
            lastItemPropertyRetrieved = item;
        }
        
        if(item==null) {
            return null;
        } else {
            return item.getItemProperty(propertyId);
        }
    }

    @Override
    public Class<?> getType(Object propertyId) {
        Class c = model.get((String) propertyId);
        if (c == null) {
            return Object.class;
        } else {
            return c;
        }
    }

    @Override
    public int indexOfId(Object itemId) {
        return getItemIdsAsList().indexOf(itemId);
    }

    @Override
    public boolean isFirstId(Object itemId) {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return false;
        } else {
            return itemId.equals(idlist.get(0));
        }
    }

    @Override
    public boolean isLastId(Object itemId) {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return false;
        } else {
            return itemId.equals(idlist.get(idlist.size() - 1));
        }
    }

    @Override
    public Object lastItemId() {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return null;
        } else {
            return idlist.get(idlist.size() - 1);
        }
    }

    @Override
    public Object nextItemId(Object itemId) {
        List idlist = getItemIdsAsList();
        int index = idlist.indexOf(itemId);
        if (index >= 0 && index < idList.size() - 1) {
            return idlist.get(index + 1);
        }
        return null;
    }

    @Override
    public Object prevItemId(Object itemId) {
        List idlist = getItemIdsAsList();
        int index = idlist.indexOf(itemId);
        if (index > 0) {
            return idlist.get(index - 1);
        }
        return null;
    }

    @Override
    public int size() {
        return getItemIdsAsList().size();
    }

    @Override
    public boolean removeItem(Object itemId) throws UnsupportedOperationException {
            repository.deleteById((IDT)itemId);
            updateIds();
            return true;
    }

    @Override
    public boolean removeContainerProperty(Object propertyId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean removeAllItems() throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * The same as calling <code>getPropertyDescriptors(beanClass, false);</code>
     */
    public static LinkedHashMap<String, Class> getPropertyDescriptors(final Class<?> beanClass) {
        return getPropertyDescriptors(beanClass, false);
    }
    
    /**
     * <p>
     * Perform introspection on a Java Bean class to find its properties.
     * </p>
     *
     * <p>
     * Note : This version only supports introspectable bean properties and
     * their getter and setter methods. Stand-alone <code>is</code> and
     * <code>are</code> methods are not supported.
     * </p>
     *
     * @param beanClass the Java Bean class to get properties for.
     * @param Whether colletions should be included in the propertyMap
     * @return an ordered map from property names to property descriptors
     */
    public static LinkedHashMap<String, Class> getPropertyDescriptors(final Class<?> beanClass, boolean excludeCollections) {
        final LinkedHashMap<String, Class> pdMap = new LinkedHashMap<String, Class>();

        // Try to introspect, if it fails, we just have an empty Item
        try {
            List<PropertyDescriptor> propertyDescriptors = getBeanPropertyDescriptor(beanClass);

            // Add all the bean properties as MethodProperties to this Item
            // later entries on the list overwrite earlier ones
            for (PropertyDescriptor pd : propertyDescriptors) {
                final Method getMethod = pd.getReadMethod();
                boolean collectionType = Collection.class.isAssignableFrom(pd.getPropertyType());
                if ((getMethod != null) && getMethod.getDeclaringClass() != Object.class && (!excludeCollections || !collectionType)) {
                    pdMap.put(pd.getName(), pd.getPropertyType());
                }
            }
        } catch (final java.beans.IntrospectionException ignored) { /* EMPTY */ }

        return pdMap;
    }

    /**
     * Returns the property descriptors of a class or an interface.
     *
     * For an interface, superinterfaces are also iterated as Introspector does
     * not take them into account (Oracle Java bug 4275879), but in that case,
     * both the setter and the getter for a property must be in the same
     * interface and should not be overridden in subinterfaces for the discovery
     * to work correctly.
     *
     * For interfaces, the iteration is depth first and the properties of
     * superinterfaces are returned before those of their subinterfaces.
     *
     * @param beanClass
     * @return
     * @throws IntrospectionException
     */
    private static List<PropertyDescriptor> getBeanPropertyDescriptor(
            final Class<?> beanClass) throws IntrospectionException {
        // Oracle bug 4275879: Introspector does not consider superinterfaces of
        // an interface
        if (beanClass.isInterface()) {
            List<PropertyDescriptor> propertyDescriptors = new ArrayList<PropertyDescriptor>();

            for (Class<?> cls : beanClass.getInterfaces()) {
                propertyDescriptors.addAll(getBeanPropertyDescriptor(cls));
            }

            BeanInfo info = Introspector.getBeanInfo(beanClass);
            propertyDescriptors.addAll(Arrays.asList(info.getPropertyDescriptors()));

            return propertyDescriptors;
        } else {
            BeanInfo info = Introspector.getBeanInfo(beanClass);
            return Arrays.asList(info.getPropertyDescriptors());
        }
    }

    public boolean isReadThrough() {
        return readThrough;
    }

    public void setReadThrough(boolean readThrough) {
        this.readThrough = readThrough;
    }

    public boolean isWriteThrough() {
        return writeThrough;
    }

    public void setWriteThrough(boolean writeThough) {
        this.writeThrough = writeThough;
    }

    public boolean isModified() {
        return false;
    }

    public void discard() throws SourceException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void commit() throws SourceException, InvalidValueException {
        throw new UnsupportedOperationException("Not supported yet.");
    }
    
}
